"use strict";
const { parseURLToResultingURLRecord } = require("../helpers/document-base-url");
const { asciiCaseInsensitiveMatch } = require("../helpers/strings");
const { navigate } = require("../window/navigation");

exports.implementation = class HTMLHyperlinkElementUtilsImpl {
  _htmlHyperlinkElementUtilsSetup() {
    this.url = null;
  }

  // https://html.spec.whatwg.org/multipage/links.html#cannot-navigate
  _cannotNavigate() {
    // TODO: Correctly check if the document is fully active
    return this._localName !== "a" && !this.isConnected;
  }

  // https://html.spec.whatwg.org/multipage/semantics.html#get-an-element's-target
  _getAnElementsTarget() {
    if (this.hasAttributeNS(null, "target")) {
      return this.getAttributeNS(null, "target");
    }

    const baseEl = this._ownerDocument.querySelector("base[target]");

    if (baseEl) {
      return baseEl.getAttributeNS(null, "target");
    }

    return "";
  }

  // https://html.spec.whatwg.org/multipage/browsers.html#the-rules-for-choosing-a-browsing-context-given-a-browsing-context-name
  _chooseABrowsingContext(name, current) {
    let chosen = null;

    if (name === "" || asciiCaseInsensitiveMatch(name, "_self")) {
      chosen = current;
    } else if (asciiCaseInsensitiveMatch(name, "_parent")) {
      chosen = current.parent;
    } else if (asciiCaseInsensitiveMatch(name, "_top")) {
      chosen = current.top;
    } else if (!asciiCaseInsensitiveMatch(name, "_blank")) {
      // https://github.com/whatwg/html/issues/1440
    }

    // TODO: Create new browsing context, handle noopener

    return chosen;
  }

  // https://html.spec.whatwg.org/multipage/links.html#following-hyperlinks-2
  _followAHyperlink() {
    if (this._cannotNavigate()) {
      return;
    }

    const source = this._ownerDocument._defaultView;
    let targetAttributeValue = "";

    if (this._localName === "a" || this._localName === "area") {
      targetAttributeValue = this._getAnElementsTarget();
    }

    const noopener = this.relList.contains("noreferrer") || this.relList.contains("noopener");

    const target = this._chooseABrowsingContext(targetAttributeValue, source, noopener);

    if (target === null) {
      return;
    }

    const url = parseURLToResultingURLRecord(this.href, this._ownerDocument);

    if (url === null) {
      return;
    }

    // TODO: Handle hyperlink suffix and referrerpolicy
    setTimeout(() => {
      navigate(target, url, {});
    }, 0);
  }

  toString() {
    return this.href;
  }

  get href() {
    reinitializeURL(this);
    const { url } = this;

    if (url === null) {
      const { href } = url;
      return href === null ? "" : href;
    }

    return url.toString();
  }

  set href(v) {
    this.setAttributeNS(null, "href", v);
  }

  get origin() {
    reinitializeURL(this);

    if (this.url === null) {
      return "";
    }

    return this.url.origin;
  }

  get protocol() {
    reinitializeURL(this);

    if (this.url === null) {
      return ":";
    }

    return this.url.scheme + ":";
  }

  set protocol(v) {
    reinitializeURL(this);

    if (this.url === null) {
      return;
    }

    this.url.protocol = v;
    updateHref(this);
  }

  get username() {
    reinitializeURL(this);
    if (this.url === null) {
      return "";
    }
    return this.url.username;
  }

  set username(v) {
    reinitializeURL(this);
    if (this.url === null) {
      return;
    }
    this.url.username = v;
    updateHref(this);
  }

  get password() {
    reinitializeURL(this);
    if (this.url === null) {
      return "";
    }
    return this.url.password;
  }

  set password(v) {
    reinitializeURL(this);
    if (this.url === null) {
      return;
    }
    this.url.password = v;
    updateHref(this);
  }

  get host() {
    reinitializeURL(this);
    if (this.url === null) {
      return "";
    }
    return this.url.host;
  }

  set host(v) {
    reinitializeURL(this);
    if (this.url === null) {
      return;
    }
    this.url.host = v;
    updateHref(this);
  }

  get hostname() {
    reinitializeURL(this);
    if (this.url === null) {
      return "";
    }
    return this.url.host;
  }

  set hostname(v) {
    reinitializeURL(this);
    if (this.url === null) {
      return;
    }
    this.url.hostname = v;
    updateHref(this);
  }

  get port() {
    reinitializeURL(this);
    if (this.url === null) {
      return "";
    }
    return this.url.port;
  }

  set port(v) {
    reinitializeURL(this);
    if (this.url === null) {
      return;
    }
    this.url.port = v;
    updateHref(this);
  }

  get pathname() {
    reinitializeURL(this);
    if (this.url === null) {
      return "";
    }

    return this.url.pathname;
  }

  set pathname(v) {
    reinitializeURL(this);
    if (this.url === null) {
      return;
    }

    this.url.pathname = v;
    updateHref(this);
  }

  get search() {
    reinitializeURL(this);
    if (this.url === null) {
      return "";
    }
    return this.url.search;
  }

  set search(v) {
    reinitializeURL(this);
    if (this.url === null) {
      return;
    }
    this.url.search = v;
    updateHref(this);
  }

  get hash() {
    reinitializeURL(this);
    if (this.url === null) {
      return "";
    }
    return this.url.hash;
  }

  set hash(v) {
    reinitializeURL(this);
    if (this.url === null) {
      return;
    }
    this.url.hash = v;
    updateHref(this);
  }
};

function reinitializeURL(hheu) {
  setTheURL(hheu);
}

function setTheURL(hheu) {
  const href = hheu.getAttributeNS(null, "href");
  if (href === null) {
    hheu.url = null;
    return;
  }

  const parsed = parseURLToResultingURLRecord(href, hheu._ownerDocument);

  hheu.url = parsed === null ? null : parsed;
}

function updateHref(hheu) {
  hheu.setAttributeNS(null, "href", hheu.url.toString());
}
