"use strict";
const DOMException = require("../generated/DOMException");
const { documentBaseURL, parseURLToResultingURLRecord } = require("../helpers/document-base-url");
const { navigate } = require("./navigation");

// Not implemented: use of entry settings object's API base URL in href setter, assign, and replace. Instead we just
// use the document base URL. The difference matters in the case of cross-frame calls.

exports.implementation = class LocationImpl {
  constructor(globalObject, args, privateData) {
    this._relevantDocument = privateData.relevantDocument;
    this.url = null;

    this._globalObject = globalObject;
  }

  get _url() {
    return this._relevantDocument._URL;
  }

  _locationObjectSetterNavigate(url) {
    // Not implemented: extra steps here to determine replacement flag.

    return this._locationObjectNavigate(url);
  }

  _locationObjectNavigate(url, { replacement = false } = {}) {
    // Not implemented: the setup for calling navigate, which doesn't apply to our stub navigate anyway.

    navigate(this._relevantDocument._defaultView, url, { replacement, exceptionsEnabled: true });
  }

  toString() {
    return this.href;
  }

  get href() {
    return this._url.href;
  }
  set href(v) {
    const newURL = URL.parse(v, { baseURL: documentBaseURL(this._relevantDocument) });
    if (newURL === null) {
      throw new TypeError(`Could not parse "${v}" as a URL`);
    }

    this._locationObjectSetterNavigate(newURL);
  }

  get origin() {
    return this._url.origin;
  }
  set origin(v) {
    const copyURL = new URL(this._url);
    copyURL.origin = v;
    this._locationObjectSetterNavigate(copyURL);
  }

  get protocol() {
    return this._url.scheme + ":";
  }
  set protocol(v) {
    const copyURL = new URL(this._url);
    copyURL.protocol = v;
    this._locationObjectSetterNavigate(copyURL);
  }

  get host() {
    return this._url.host;
  }
  set host(v) {
    const copyURL = new URL(this._url);
    copyURL.host = v;
    this._locationObjectSetterNavigate(copyURL);
  }

  get hostname() {
    return this._url.hostname;
  }
  set hostname(v) {
    const copyURL = new URL(this._url);
    copyURL.hostname = v;
    this._locationObjectSetterNavigate(copyURL);
  }

  get port() {
    return this._url.port;
  }
  set port(v) {
    const copyURL = new URL(this._url);
    copyURL.port = v;
    this._locationObjectSetterNavigate(copyURL);
  }

  get pathname() {
    return this._url.pathname;
  }
  set pathname(v) {
    const copyURL = new URL(this._url);
    copyURL.pathname = v;
    this._locationObjectSetterNavigate(copyURL);
  }

  get search() {
    return this._url.search;
  }
  set search(v) {
    const copyURL = new URL(this._url);
    copyURL.search = v;
    this._locationObjectSetterNavigate(copyURL);
  }

  get hash() {
    return this._url.hash;
  }
  set hash(v) {
    const copyURL = new URL(this._url);
    copyURL.hash = v;
    this._locationObjectSetterNavigate(copyURL);
  }

  assign(url) {
    // Should be entry settings object; oh well
    const parsedURL = parseURLToResultingURLRecord(url, this._relevantDocument);

    if (parsedURL === null) {
      throw DOMException.create(this._globalObject, [
        `Could not resolve the given string "${url}" relative to the base URL "${this._relevantDocument.URL}"`,
        "SyntaxError"
      ]);
    }

    this._locationObjectNavigate(parsedURL);
  }

  replace(url) {
    // Should be entry settings object; oh well
    const parsedURL = parseURLToResultingURLRecord(url, this._relevantDocument);

    if (parsedURL === null) {
      throw DOMException.create(this._globalObject, [
        `Could not resolve the given string "${url}" relative to the base URL "${this._relevantDocument.URL}"`,
        "SyntaxError"
      ]);
    }

    this._locationObjectNavigate(parsedURL, { replacement: true });
  }

  reload() {
    const flags = { replace: true, reloadTriggered: true, exceptionsEnabled: true };
    navigate(this._relevantDocument._defaultView, this._url, flags);
  }
};
